function [mloglik,se2,F,pe,A,Pa,setup,pred,smooth] = loglik(m,data,Zd,setup,varargin)
%
% LOGLIK  Use Kalman filter to evaluate likelihood and to project state vector on observables.
%
% Syntax:
%   [mloglik,se2,F,pe] = loglik(m,data,Z,setup,...)
%   [mloglik,se2,F,pe,A,Pa,setup] = loglik(m,data,Z,setup,...)
%   [mloglik,se2,F,pe,A,Pa,setup,pred,smooth] = loglik(m,data,Z,setup,...)
% Arguments:
%   mloglik numeric; se2,F,pe,A,Pa numeric; setup cell; pred struct; smooth struct; m model; data cell|numeric; Z numeric
% Options:
%   'exclude' cellstr (empty) List of measurement variables to be treated as deterministic (excluded from likelihood function)
%   'precision' char (get(m,'precision')) Numerical precision of prediction and smoothing datapacks
%   'relative' logical (true) If true MSE matrices are scaled by estimated variance factor
%
% The IRIS Toolbox 2007/02/03. Copyright 2007 Jaromir Benes.

% no deterministic regressors
if nargin < 3, Zd = []; end

% user defined setup matrices
usersetup = nargin > 3 && ~isempty(setup);

default = {...
  'exclude',{},...
  'precision',m.precision,...
  'relative',true,...
};
options = passopt(default,varargin{:});

% function body ---------------------------------------------------------------------------------------------

if isempty(m.solution) || isempty(m.solution{1}), error_(12); end

realsmall = getrealsmall();
[ny,nx,nf,nb,ne,np,nalt] = dim_(m);
me = [];
if iscell(data)
  range = data{4};
  if length(data) > 4, me = data{5}; end
  data = data{1};
  nper = size(data,2);
else
  data = transpose(data);
  nper = size(data,2);
  range = 1 : nper;
end
ndata = size(data,3);  
nloop = max([ndata,nalt]);

% number of deterministic regressors
nd = size(Zd,2);
% expand Zd to match the number of periods if necessary
if nd > 0 && size(Zd,3) == 1 && nper > 1, Zd = Zd(:,:,ones([1,nper])); end

mloglik = nan([1,nloop]);
se2 = nan([1,nloop]);

ixdeterm = false([ny,1]);
if ~isempty(options.exclude) && (iscellstr(options.exclude) | ischar(options.exclude))
  aux = findnames(m.name(m.nametype == 1),options.exclude);
  if any(isnan(aux))
    warning_(1,options.exclude(isnan(aux)));
    aux(isnan(aux)) = [];  
  end
  ixdeterm(aux) = true;
end

if nargout > 7
  pred.mean = {nan([ny,nper,nloop],options.precision),nan([nx,nper,nloop],options.precision),zeros([ne,nper,nloop],options.precision),range};
  pred.mse = {nan([ny,ny,nper,nloop],options.precision),nan([nx,nx,nper,nloop],options.precision),[],range};
  smooth.mean = {nan([ny,nper,nloop],options.precision),nan([nx,nper,nloop],options.precision),nan([ne,nper,nloop],options.precision),range};
  smooth.mse = {zeros([ny,ny,nper,nloop],options.precision),nan([nx,nx,nper,nloop],options.precision),[],range};
  if ~isempty(me)
    pred.mean{5} = me;
    smooth.mean{5} = me;
  end
end

mloglik = nan([1,nloop]);
se2 = nan([1,nloop]);

% get index of available solutions
[flag,nansol] = isnan(m,'solution');

for iloop = 1 : nloop

  if iloop <= nalt && nansol(iloop) == true, continue, end
  if iloop > nalt && nansol(end) == true, continue, end

  if iloop <= ndata, y = data(:,:,iloop); end

  if iloop <= nalt
    % set up filter matrices
    [T,R,K,Z,H,D,Za,Omega] = sspace_(m,iloop);
    % separate measurement and transition errors
    ixte = any(R,1);
    ixme = any(H,1);
    R = R(:,ixte);
    H = H(:,ixme);
    Omegate = Omega(ixte,ixte);
    Omegame = Omega(ixme,ixme);
    % initial distribution of transformed predetermined variables
    if usersetup == true
      % reuse data-independent user setup matrices
      ixunit = setup{1};
      ixfixed = setup{2};
      x00 = setup{3};
      P00 = setup{4};
    else
      [ixunit,ixfixed,x00,P00] = properinit_();
      if nargout > 6, setup = {ixunit,ixfixed,x00,P00}; end
    end
    if nd > 0
      T(end+1:end+nd,end+1:end+nd) = eye(nd);
      R(end+1:end+nd,:) = 0;
      K(end+1:end+nd,:) = 0;
      x00(end+1:end+nd,:) = 0;
      P00(end+1:end+nd,end+1:end+nd) = 0;
      if size(Zd,3) == 1, Z = [Z,Zd];
        else, Z = [Z(:,:,ones([1,nper])),Zd]; end 
      ixunit(end+1:end+nd) = true;
      ixfixed(end+1:end+nd) = true;
    end
    if size(Z,3) == 1, Z = Z(:,:,ones([1,nper])); end
    if nargout > 7 % predicted or smoothed series requested
      ixunit = [false([1,nf]),ixunit];
      ixfixed = [false([1,nf]),ixfixed];
      T = [zeros([nx+nd,nf]),T];
      Z = [zeros([ny,nf,nper]),Z];
      x00 = [zeros([nf,1]);x00];
      P00 = blkdiag(zeros(nf),P00);
    else % predicted or smoothed series not requested
      % remove non-predetermined variables
      T = T(nf+1:end,:);
      R = R(nf+1:end,:);
      K = K(nf+1:end,:);
      % remove unit-root states that are not projected on observables
      ixremove = ixunit & ~ixfixed;
      if any(ixremove)
        T = T(~ixremove,~ixremove);
        R = R(~ixremove,:);
        K = K(~ixremove,:);
        Z = Z(:,~ixremove,:);
        x00 = x00(~ixremove);
        P00 = P00(~ixremove,~ixremove);
        ixunit = ixunit(~ixremove);
        ixfixed = ixfixed(~ixremove);
      end
    end
    Sigmate = R*Omegate*transpose(R);
    Sigmame = H*Omegame*transpose(H);
    ixfilter = any(any(abs(Z) > realsmall,3),1);
    ixstate = any(abs(T) > realsmall,1);
    if usersetup == true
      % reuse data-independent user setup matrices
      P10 = setup{5};
      F = setup{6};
      F_Z = setup{7};
      P10_Zt = setup{8};
      L = setup{9};
      M = setup{10};
      Q = setup{11};
      Qt_Zt = setup{12};
      U = setup{13};
    else
      % evaluate data-independent information
      [P10,F,F_Z,P10_Zt,L,M,Q,Qt_Zt,U] = setup_();
      if nargout > 6, setup = [setup,{P10,F,F_Z,P10_Zt,L,M,Q,Qt_Zt,U}]; end
    end
  end

  % prediction error decomposition
  [x10,pe,F_pe,A] = filter_();
  % correct for fixed initial conditions
  if any(ixfixed), correct_(); end
  % compute minus loglikelihood and variance scale
  [mloglik(iloop),se2(iloop)] = loglik_();

  if nargout > 7
    % backward smoothing
    [x1T,P1T,em,et] = smooth_();
    % create new copy as P10 must be preserved
    x10_ = x10(1:nx,:);
    x1T_ = x1T(1:nx,:);
    P10_ = P10(1:nx,1:nx,:);
    P1T_ = P1T(1:nx,1:nx,:);
  end
  if nargout > 4 && nd > 0
    % parameters concentrated out of loglik function
    A = A(end-nd+1:end);
    if nargout > 5
      Pa = inv(U);
      Pa = Pa(end-nd+1:end,end-nd+1:end);
    end
  else
    A = zeros([0,1]);
    Pa = [];
  end

  if options.relative == true, scale_(); end

  if nargout > 7
    % tranformation of predetermined states
    transform_();
    % prediction data pack
    pred.mean{1}(:,:,iloop) = y10;
    pred.mean{2}(:,:,iloop) = x10_;
    pred.mse{1}(:,:,:,iloop) = F;
    pred.mse{2}(:,:,:,iloop) = P10_;
    % smoothing data pack
    smooth.mean{1}(:,:,iloop) = y;
    smooth.mean{2}(:,:,iloop) = x1T_;
    smooth.mean{3}(ixte,:,iloop) = et;
    smooth.mean{3}(ixme,:,iloop) = em;
    smooth.mse{2}(:,:,:,iloop) = P1T_;
  end % of if

end % of for iloop

% end of function body --------------------------------------------------------------------------------------

  function transform_() % nested function -------------------------------------------------------------------
  if isempty(Za), return, end
  x10_(nf+1:end,:) = Za*x10_(nf+1:end,:);
  x1T_(nf+1:end,:) = Za*x1T_(nf+1:end,:);
  for t = 1 : nper
    P10_(nf+1:end,:,t) = Za*P10_(nf+1:end,:,t);
    P10_(:,nf+1:end,t) = P10_(:,nf+1:end,t)*transpose(Za);
    P1T_(nf+1:end,:,t) = Za*P1T_(nf+1:end,:,t);
    P1T_(:,nf+1:end,t) = P1T_(:,nf+1:end,t)*transpose(Za);
  end
  end % of nested function ----------------------------------------------------------------------------------

  function [ixunit,ixfixed,a00,P00] = properinit_() % subfunction -------------------------------------------
  TT = T(nf+1:end,:);
  KK = K(nf+1:end,1);
  RR = R(nf+1:end,:);
  a00 = zeros([nb,1]);
  P00 = zeros(nb);
  [P00,ixunit] = acovf(TT,RR,[],[],[],[],[],Omegate,m.eigval(1,:,iloop),0);
  P00(ixunit,:) = 0;
  P00(:,ixunit) = 0;
  if any(~ixunit), a00(~ixunit,1) = (eye(sum(~ixunit)) - TT(~ixunit,~ixunit)) \ KK(~ixunit); end
  % do not estimate initial conditions of unobservables for which no information is in observables
  ixfixed = ixunit;
  % if model is not first-difference stationary then the second term can be binding
  index = all(abs(Z(:,ixfixed)) < realsmall,1) & all(abs(TT(:,ixfixed) - eye([nb,sum(ixfixed)])) < realsmall,1);
  ixfixed(index) = false;
  end % of nested function ----------------------------------------------------------------------------------

  function [P10,F,F_Z,P10_Zt,L,M,Q,Qt_Zt,U] = setup_() % nested function ------------------------------------
  nx_ = size(T,1);
  P10 = zeros([nx_,nx_,nper]);
  F = zeros([ny,ny,nper]);
  F_Z = zeros([ny,nx_,nper]);
  P10_Zt = zeros([nx_,ny,nper]);
  L = zeros([nx_,nx_,nper]);
  M = zeros([nx_,nx_,nper]);
  % Q*A = extra effect of estimated initial condition on state vector
  Q = zeros([nx_,sum(ixfixed),nper]);
  Qt_Zt = zeros([sum(ixfixed),ny,nper]);
  U = 0;
  eyeindex = find(eye(size(T)));
  for t = 1 : nper
    if t == 1
      P10(:,:,t) = T(:,~ixfixed)*P00(~ixfixed,~ixfixed)*transpose(T(:,~ixfixed)) + Sigmate;
      Q(:,:,t) = T(:,ixstate)*ref(eye(nx_),ixstate,ixfixed);
    else
      P10(:,:,t) = T(:,ixstate)*P11(ixstate,ixstate)*transpose(T(:,ixstate)) + Sigmate;
      Q(:,:,t) = M(:,:,t-1)*Q(:,:,t-1);
    end
    ixobserved = vech(~isnan(y(:,t)));
    P10_Zt(:,ixobserved,t) = P10(:,ixfilter,t)*transpose(Z(ixobserved,ixfilter,t));
    F(ixobserved,ixobserved,t) = Z(ixobserved,ixfilter,t)*P10_Zt(ixfilter,ixobserved,t) + Sigmame(ixobserved,ixobserved);
    F_Z(ixobserved,ixfilter,t) = F(ixobserved,ixobserved,t)\Z(ixobserved,ixfilter,t);
    L(:,ixfilter,t) = P10_Zt(:,ixobserved,t)*F_Z(ixobserved,ixfilter,t);
    M(:,:,t) = T - T(:,ixstate)*L(ixstate,:,t);
    P11(ixstate,ixstate) = P10(ixstate,ixstate,t) - L(ixstate,:,t)*P10(:,ixstate,t);
    if any(ixobserved) && any(ixfixed)
      Qt_Zt(:,ixobserved,t) = transpose(Z(ixobserved,:,t)*Q(:,:,t));
      U = U + Qt_Zt(:,ixobserved,t)*F_Z(ixobserved,ixfilter,t)*Q(ixfilter,:,t);
    end
  end
  end % of nested function ----------------------------------------------------------------------------------

  function [x10,pe,F_pe,A] = filter_() % nested function ----------------------------------------------------
  x10 = zeros([size(T,1),nper]);
  x11 = zeros([size(T,1),nper]);
  y10 = zeros([size(Z,1),nper]);
  pe = zeros([size(Z,1),nper]);
  F_pe = zeros([ny,nper]);
  V = 0;
  Logl1 = 0;
  % prediction and updating
  for t = 1 : nper
    if t == 1, x10(:,t) = T(:,~ixfixed)*x00(~ixfixed,1) + K;
      else, x10(:,t) = T(:,ixstate)*x11(ixstate,t-1) + K; end
    ixobserved = vech(~isnan(y(:,t)));
    y10(ixobserved,t) = Z(ixobserved,ixfilter,t)*x10(ixfilter,t) + D(ixobserved,:);
    pe(ixobserved,t) = y(ixobserved,t) - y10(ixobserved,t);
    F_pe(ixobserved,t) = F(ixobserved,ixobserved,t)\pe(ixobserved,t);
    x11(:,t) = x10(:,t) + P10_Zt(:,:,t)*F_pe(:,t);
    if any(ixobserved) && any(ixfixed), V = V + Qt_Zt(:,ixobserved,t)*F_pe(ixobserved,t); end
  end
  % estimate fixed initial conditions
  lastwarn('');
  warning('off','MATLAB:SingularMatrix');
  warning('off','MATLAB:nearlySingularMatrix');
  if any(ixfixed), A = U \ V;
    else A = zeros([0,1]); end
  [warnmessage,warnid] = lastwarn();
  if ~isempty(warnid) && strcmp(warnid,'MATLAB:nearlySingularMatrix') || strcmp(warnid,'MATLAB:SingularMatrix')
    warning_(28);
  end
  warning('on','MATLAB:SingularMatrix');
  warning('on','MATLAB:nearlySingularMatrix');
  end % of nested function ----------------------------------------------------------------------------------
  
  function correct_() % nested function ---------------------------------------------------------------------
  for t = 1 : nper
    ixobserved = vech(~isnan(y(:,t)));
    Q_A = Q(:,:,t)*A;
    x10(:,t) = x10(:,t) + Q_A;
    y10(ixobserved,t) = y10(ixobserved,t) + Z(ixobserved,ixfilter,t)*Q_A(ixfilter,:);
    pe(ixobserved,t) = y(ixobserved,t) - y10(ixobserved,t);
    F_pe(ixobserved,t) = F_pe(ixobserved,t) - F_Z(ixobserved,ixfilter,t)*Q_A(ixfilter,:);
  end
  end % of nested function ----------------------------------------------------------------------------------

  function [mloglik_,se2_] = loglik_() % nested function ----------------------------------------------------
  L0 = 0;
  L1 = 0;
  nobs = 0;
  for t = 1 : nper
    ixobserved = vech(~isnan(y(:,t)) & ~ixdeterm);
    nobs = nobs + sum(ixobserved);
    L0 = L0 + log(det(F(ixobserved,ixobserved,t)));
    L1 = L1 + transpose(pe(ixobserved,t))*F_pe(ixobserved,t);
  end
  % estimate scaling factor for variances-covariances
  if options.relative == true, se2_ = L1 / nobs;
    else, se2_ = 1; end
  % minus log-likelihod
  mloglik_ = 1/2*(nobs*log(2*pi) + (L0 + nobs*log(se2_)) + L1/se2_);
  end % of nested function ----------------------------------------------------------------------------------

  function [x1T,P1T,em,et] = smooth_() % nested function ----------------------------------------------------
  nx_ = size(T,1);
  x1T = zeros([nx_,nper]);
  P1T = zeros([nx_,nx_,nper]);
  em = zeros([sum(ixme),nper]);
  et = zeros([sum(ixte),nper]);
  b = zeros([nx_,1]);
  N = zeros(nx_);
  for t = nper : -1 : 1
    ixobserved = ~isnan(y(:,t));
    em(:,t) = Omegame*transpose(H(ixobserved,:))*(F_pe(ixobserved,t) - F_Z(ixobserved,:,t)*transpose(T(:,ixstate)*P10(ixstate,:,t))*b);
    b = transpose(Z(ixobserved,:,t))*F_pe(ixobserved,t) + transpose(M(:,:,t))*b;
    et(:,t) = Omegate*transpose(R)*b;
    N = transpose(Z(ixobserved,:,t))*F_Z(ixobserved,:,t) + transpose(M(:,:,t))*N*M(:,:,t);
    x1T(:,t) = x10(:,t) + P10(:,:,t)*b;
    P1T(:,:,t) = P10(:,:,t) - P10(:,:,t)*N*P10(:,:,t);
  end
  end % of nested function ----------------------------------------------------------------------------------

  function scale_() % nested function -----------------------------------------------------------------------
  F = F*se2(iloop);
  P10 = P10*se2(iloop);
  if nargout > 6, P1T = P1T*se2(iloop); end
  if nargout > 5 && nd > 0, Pa = Pa*se2(iloop); end
  end % of nested function ----------------------------------------------------------------------------------

end % of primary function -----------------------------------------------------------------------------------