function [m,assign] = parser_(m,assign)
%
% MODEL/PRIVATE/PARSER_  Parse model code file.
%
% The IRIS Toolbox 2007/02/03. Copyright 2007 Jaromir Benes.

% read in either text file or text string
text = readcode_(m.fname,assign);

% desired versions of equations
%if isnumeric(version), version = num2str(version); end
%if ~isempty(version), text = regexprep(text,sprintf('#%s[\\s,:]+',version),''); end
%text = regexprep(text,'#(.*?);','');

% obsolete or mutliple syntax
text = strrep(text,'@ifversion','@if');
text = strrep(text,'@coefficients','@parameters');
text = strrep(text,'@variables:innovation','@variables:residual');
text = strrep(text,'@residuals','@variables:residual');
text = strrep(text,'@nonlinear:','');
text = strrep(text,'@nonlinear','');

% check for invalid model keywords (i.e. @...)
keywords = regexp(text,'@[\w:]*','match');
invalid = chkkey_(keywords);
if ~isempty(invalid), error_(35,invalid,m.fname); end % invalid keywords

% linear model declaration
aux = length(text);
text = strrep(text,'@linear','');
if aux > length(text), m.linear = true; end

% remove redundant comments
ltext = Inf;
while length(text) < ltext
  ltext = length(text);
  text = regexprep(text,'"[^"]*"[^\w"]*("[^"]*")','$1');
end

% read blocks
blkname = {...
  '@variables:measurement',...     % 1
  '@variables:transition',...      % 2
  '@variables:residual',...        % 3
  '@parameters',...                % 4
  '@variables.log',...             % 5
  '@equations:measurement',...     % 6
  '@equations:transition',...      % 7
  '@substitutions',...             % 8
  '@outside',...                   % 9
  '@mapoftunes',...                % 10
};

nblk = length(blkname);

eoblk = sprintf('|%s',blkname{:});
eoblk = sprintf('(?=$%s)',eoblk);

% read blocks of declarations
block = cell([1,nblk]);
for i = 1 : length(blkname)
  patt = [blkname{i},'(.*?)',eoblk];
  tokens = regexp(text,patt,'tokens');
  tokens = [tokens{:}];
  if ~isempty(tokens)
    % @allbut must be in all or none of log declaration blocks
    if i == 5, chkallbut_(tokens); end
    block{i} = regexprep([tokens{:}],'\n',' ');
  else
    block{i} = '';
  end
end

% variables and parameters ----------------------------------------------------------------------------------

% read individual names of variables and parameters
for i = 1 : 4
  [name,label,value] = namelist_(block{i});
  m.name = [m.name,name];
  m.namelabel = [m.namelabel,label];
  m.nametype = [m.nametype,i*ones(size(name),'uint8')];
  if i ~= 3 % do not assign steady state of residuals
    % assign a value from declaration if not in the assign database assign
    index = find(~cellfun(@isempty,value));
    for j = index
      if ~isfield(assign,name{j}), assign.(name{j}) = str2num(value{j}); end
    end
  end
end
m.namelabel = strrep(m.namelabel,'"','');

% take out std.error parameters if declared by user
index = strmatch('std_',m.name);
if ~isempty(index)
  m.name(index) = [];
  m.namelabel(index) = [];
  m.nametype(index) = [];
end

% add std.err. parameters
index = m.nametype == 3;
n = sum(index);
m.name(end+(1:n)) = strrep('std_$','$',m.name(index));
m.nametype(end+(1:n)) = 4;
tmp = m.namelabel(index);
try, nonempty = ~cellfun(@isempty,tmp);
  catch, nonempty = ~cellfun('isempty',tmp); end
for i = find(nonempty), tmp{i} = sprintf('SE of %s',tmp{i}); end
m.namelabel(end+(1:n)) = tmp;

% read substitutions
patt = '(\w+)\s*=\s*(.*?)\s*;';
tokens = regexp(block{8},patt,'tokens');
aux = [tokens{:}];
n = length(aux)/2;
subsname(1:n) = {''};
subsbody(1:n) = {''};
if ~isempty(aux)
  subsname = aux(1:2:end);
  subsbody = aux(2:2:end);  
end

chklist = [m.name,subsname];

% leading characters of names and substitutions must not be 0-9 or _
patt = '^[0-9_]';
match = regexp(chklist,patt,'match');
try, aux = ~cellfun(@isempty,match);
  catch, aux = ~cellfun('isempty',match); end
if any(aux), error_(2,chklist(aux),m.fname); end

% multiple declarations of names and substitutions
[check,index] = sort(chklist);
aux = all(diff(double(char(check)),1,1) == 0,2);
if any(aux), error_(3,chklist(sort(index(aux))),m.fname); end

% read log declarations
m.log = false(size(m.name));
m.log(m.nametype == 1 | m.nametype == 2) = iff(isempty(strfind(block{5},'@allbut')),false,true);
block{5} = strrep(block{5},'@allbut','');
logname = namelist_(block{5});
logname = unique(logname);
list = m.name(m.nametype == 1 | m.nametype == 2);
invalid = cell([1,0]);
for i = 1 : length(logname)
  aux = strmatch(logname{i},list,'exact');
  if ~isempty(aux) && (m.nametype(aux) == 1 || m.nametype(aux) == 2)
    m.log(aux) = ~m.log(aux);
  else
    invalid{end+1} = logname{i};
  end
end
if ~isempty(invalid)
  error_(4,invalid,m.fname);
end

m.ixname = cell([1,4]);
m.nname = zeros([1,4]);
for i = 1 : 4
  m.ixname{i} = find(m.nametype == i);
  m.nname(i) = length(m.ixname{i});
end

% add substitutions for parameters (hard number strings)
chk = [block{6:7}];
for i = find(m.nametype == 4)
  if isempty(strfind(chk,['$',m.name{i},'$'])), continue, end
  subsname{end+1} = m.name{i};
  try, tmp = assign.(m.name{i})(1);
    catch tmp = NaN; end
  subsbody{end+1} = sprintf('%.16g',tmp);
end

offset = 255;
namecode = char(offset+(1:length(m.name)));
subscode = char(offset+length(m.name)+(1:length(subsname)));

% external blocks -------------------------------------------------------------------------------------------

% outside equations
tokens = regexp(strtrim(block{9}),'(?<label>"[^"]*")?\s*(?<lhs>\w+)\s*=\s*(?<rhs>[^;]*)\s*;','names');
m.outside.label = {tokens.label};
m.outside.lhs = {tokens.lhs};
m.outside.rhs = {tokens.rhs};
% replac x{-1} with d.x{-1}(t,:)
patt = '(?<![\w@])([a-zA-Z]\w*(\{[\+-]?\d*\})?)';
repl = 'd.$1(t,:)';
m.outside.rhs = regexprep(m.outside.rhs,patt,repl);
m.outside.rhs = strrep(m.outside.rhs,'@','');
m.outside.rhs = strrep(m.outside.rhs,'*','.*');
m.outside.rhs = strrep(m.outside.rhs,'/','./');
m.outside.rhs = strrep(m.outside.rhs,'^','.^');

% map of tunes
m.mapoftunes = regexp(strtrim(block{10}),'(?<variable>\w+)\s*~\s*(?<residual>\w+)','names');
variable = {m.mapoftunes.variable};
residual = {m.mapoftunes.residual};
% invalid variable names
index = findnames(m.name(m.nametype <= 2),variable);
variableinvalid = variable(isnan(index));
if ~isempty(variableinvalid), error_(49,variableinvalid); end
% invalid shock names
index = findnames(m.name(m.nametype == 3),residual);
residualinvalid = residual(isnan(index));
if ~isempty(residualinvalid), error_(50,residualinvalid); end

% equations -------------------------------------------------------------------------------------------------

% remove extra ;s from equations
patt = '(\s*;){2,}';
block(6:7) = regexprep(block(6:7),patt,';');
block(6:7) = strrep(block(6:7),'@lineareqtn','@lineareqtn:');

% read individual equations
patt = '\s*(@lineareqtn:)?\s*("[^"]*")?([^"]*?)(=[^"]*?)?;';
[eqtn,tokens] = regexp(block(6:7),patt,'match','tokens');
m.eqtntype = [1*ones(size(eqtn{1}),'uint8'),2*ones(size(eqtn{2}),'uint8')];
m.eqtn = [eqtn{:}];
tokens = [tokens{:}];
for j = 1 : length(tokens)
  m.eqtnF{end+1} = iff(isempty(tokens{j}{4}),sprintf('%s;',tokens{j}{3}),sprintf('(%s)-(%s);',tokens{j}{3},tokens{j}{4}(2:end))); % drop leadign equal sign
  m.eqtnlabel{end+1} = tokens{j}{2};
  if m.linear == false, m.nonlin(end+1) = isempty(tokens{j}{1});
    else, m.nonlin(end+1) = false; end
end

% replace [] brackets with ()
m.eqtnF = strrep(m.eqtnF,'[','(');
m.eqtnF = strrep(m.eqtnF,']',')');

% remove blank spaces
m.eqtn = regexprep(m.eqtn,{'\s+','".*?"'},{'',''});
m.eqtnF = regexprep(m.eqtnF,'\s+','');
m.eqtnlabel = strrep(m.eqtnlabel,'"','');

% replace names with code characters
[patt,repl] = deal(cell([1,0]));
for i = 1 : length(subsname)
  patt{end+1} = ['\$',subsname{i},'(\{.*?\})?\$'];
  repl{end+1} = [subscode(i),'$1'];
end
try, len = cellfun(@length,m.name);
  catch, len = cellfun('length',m.name); end
[aux,index] = sort(len,2,'descend');
for i = index
  patt{end+1} = ['\<',m.name{i},'\>'];
  repl{end+1} = namecode(i);
end
m.eqtnF = regexprep(m.eqtnF,patt,repl);
subsbodyF = subsbody;
subsbodyF = regexprep(subsbodyF,patt,repl);

% add implicit zero time subscripts in F equations and substitutions
if ~isempty(namecode) || ~isempty(subscode)
  startcode = char(min(double([namecode,subscode])));
  endcode = char(max(double([namecode,subscode])));
  patt = ['([',startcode,'-',endcode,'])(?!\{)'];
  repl = '$1{0}';
  m.eqtnF = regexprep(m.eqtnF,patt,repl);
  subsbodyF = regexprep(subsbodyF,patt,repl);
end

% add shifts to subscripts in F substitutions
% make substitutions in substitutions
% make substitutions in F equations
pattF = {};
pattF(1:length(subsname)) = {''};
replF = {};
replF(1:length(subsname)) = {''};
for i = 1 : length(subsname)
  patt0 = '\{(.*?)\}';
  repl0 = '\{$1+(\$1)\}';
  pattF{i} = [subscode(i),'\{(.*?)\}'];
  replF{i} = regexprep(subsbodyF{i},patt0,repl0);
  subsbodyF = regexprep(subsbodyF,pattF,replF);
end
m.eqtnF = regexprep(m.eqtnF,pattF,replF);

% evaluate time subscripts in F equations
% replace {} with ()
invalid = {};
patt0 = '\{(.*?)\}';
for i = 1 : length(m.eqtnF)
  tokens = regexp(m.eqtnF{i},patt0,'tokens');
  [patt,repl] = deal(cell(size(tokens)));
  if ~isempty(patt)
    [patt{:}] = deal(patt0);
    aux = vech(str2num(char([tokens{:}])));
    if length(aux) ~= length(tokens) || any(isnan(aux)) || any(isinf(aux))
      invalid{end+1} = m.eqtn{i};
      continue;
    end
    for j = 1 : length(tokens)
      repl{j} = sprintf(iff(aux(j) >= 0,'(t+%g)','(t-%g)'),abs(aux(j)));
    end
    m.eqtnF{i} = regexprep(m.eqtnF{i},patt,repl,'once');
  end
end

if ~isempty(invalid)
  error_(9,invalid,m.fname);
end

% replace matrix operators
m.eqtnF = strrep(m.eqtnF,'*','.*');
m.eqtnF = strrep(m.eqtnF,'/','./');
m.eqtnF = strrep(m.eqtnF,'^','.^');

% math functions
m.eqtnF = strrep(m.eqtnF,'@','');
m.eqtnS = m.eqtnF;

% replace variables with x vector in full equations
patt = {};
patt(1:length(m.name)) = {''};
replF = {};
replF(1:length(m.name)) = {''};
replS = {};
replS(1:length(m.name)) = {''};
for i = 1 : length(m.name)
  patt{i} = [namecode(i),'\(t([\+\-]\d+)\)'];
  % remove time shifts from parameters
  if m.nametype(i) == 4, replF{i} = sprintf('x(:,%d,t+0)',i);
    else, replF{i} = sprintf('x(:,%d,t$1)',i); end
  if m.nametype(i) == 3
    replS{i} = '0';
  else
    replS{i} = sprintf('(x(%d)+($1)*dx(%d))',i,i);    
    if m.nametype(i) <= 2 && m.log(i) == true
      replS{i} = sprintf('exp(%s)',replS{i});
    end
  end
end
m.eqtnF = regexprep(m.eqtnF,patt,replF);
m.eqtnS = regexprep(m.eqtnS,patt,replS);

% remove zero shifts from S equations
m.eqtnS = regexprep(m.eqtnS,'+\(\+0\)\*dx\(\d+\)','');
m.eqtnS = strrep(m.eqtnS,'exp(0)','1');

% occurences
patt = 'x\(:,(\d+),t([\+\-](\d+))\)';
maxT = 0;
minT = 0;
occur = deal(cell([1,length(m.eqtnF)]));
for i = 1 : length(m.eqtnF)
  tokens = regexp(m.eqtnF{i},patt,'tokens');
  tokens = str2num(char([tokens{:}]));
  occur{i} = reshape(tokens,[2,length(tokens)/2]);
  maxT = max([maxT,occur{i}(2,:)]);
  minT = min([minT,occur{i}(2,:)]);    
end
maxT = maxT + 1;
minT = minT - 1;
t = 1 - minT;
m.tzero = t;
n = sum(m.nametype <= 4);
nT = maxT - minT + 1;
m.occur = false([length(m.eqtnF),n,nT]);
for i = 1 : length(m.eqtnF)
  index = sub2ind(size(m.occur),i*ones([1,size(occur{i},2)]),occur{i}(1,:),t+occur{i}(2,:));
  m.occur(index) = true;
end

m.ixeqtn = cell([1,2]);
m.neqtn = zeros([1,4]);
for i = 1 : 2
  m.ixeqtn{i} = find(m.eqtntype == i);
  m.neqtn(i) = length(m.ixeqtn{i});
end

chksyntax_(m,nT,t);
chkstructure_();

  function chkstructure_() % nested function ----------------------------------------------------------------

  %% one transition variable at least
  if m.nname(2) == 0
    error_(6,{},'No transition variable.',m.fname);
  end
  
  %% current dates of all transition variables
  aux = find(any(m.occur(m.ixeqtn{2},m.ixname{2},t),1) == 0);
  if ~isempty(aux)
    error_(6,m.name(1,m.ixname{2}(aux)),'Missing current date of transition variable ''%s''.',m.fname);
  end
  
  %% current dates of all measurement variables
  aux = find(any(m.occur(m.ixeqtn{1},m.ixname{1},t),1) == 0);
  if ~isempty(aux)
    error_(6,m.name(1,m.ixname{1}(aux)),'Missing current date of measurement variable: ''%s''.',m.fname);
  end

  %% # measurement equations = # measurement variables
  if m.nname(1) ~= m.neqtn(1)
    error_(6,{m.neqtn(1),m.nname(1)},'%g measurement equation(s) for %g measurement variable(s).',m.fname);
  end

  %% # transition equations = # transition variables
  if m.nname(2) ~= m.neqtn()
    error_(6,{m.neqtn(2),m.nname(2)},'%g transition equation(s) for %g transition variable(s).',m.fname);
  end
    
  %% no lags/leads of measurement variables
  aux = find(any(any(m.occur(:,m.ixname{1},[1:t-1,t+1:end]),3),1));
  if ~isempty(aux)
    error_(6,m.name(1,m.ixname{1}(aux)),'Lag(s) or lead(s) of measurement variable: ''%s''.',m.fname);
  end
  
  %% no lags/leads of residual variables
  aux = find(any(any(m.occur(:,m.ixname{3},[1:t-1,t+1:end]),3),1));
  if ~isempty(aux)
    error_(6,m.name(1,m.ixname{3}(aux)),'Lag(s) or lead(s) of residual variable: ''%s''.',m.fname);
  end
  
  %% no lags/leads of parameters
  aux = find(any(any(m.occur(:,m.ixname{4},[1:t-1,t+1:end]),3),1));
  if ~isempty(aux)
    error_(6,m.name(1,m.ixname{4}(aux)),'Lag(s) or lead(s) of parameter: ''%s''.',m.fname);
  end
  
  %% no measurement variables in transition equations
  aux = find(any(any(m.occur(m.ixeqtn{2},m.ixname{1},:),3),2));
  if ~isempty(aux)
    error_(6,m.eqtn(1,m.ixeqtn{2}(aux)),'Transition equation contains measurement variable(s): ''%s''.',m.fname);
  end
  
  %% no leads of transition variables in measurement equations
  aux = find(any(any(m.occur(m.ixeqtn{1},m.ixname{2},t+1:end),3),2));
  if ~isempty(aux)
    error_(6,m.eqtn(1,m.ixeqtn{1}(aux)),'Lead(s) of transition variable(s) in measurement equation: %s',m.fname);
  end
    
  %% current date of any transition variable in each transition equation
  aux = find(any(m.occur(m.ixeqtn{2},m.ixname{2},t),2) == 0);
  if ~isempty(aux)
    error_(6,m.eqtn(1,m.ixeqtn{2}(aux)),'No current-dated transition variable in transition equation: %s.',m.fname);
  end
  
  %% current date of any measurement variable in each measurement equation
  aux = find(any(m.occur(m.ixeqtn{1},m.ixname{1},t),2) == 0);
  if ~isempty(aux)
    error_(6,m.eqtn(1,m.ixeqtn{1}(aux)),'No current-dated measurement variable in measurement equation: %s',m.fname);
  end
  
  %% no residual variable simultaneously in transition and measurement equations
  aux1 = any(m.occur(m.ixeqtn{1},m.ixname{3},t),1);
  aux2 = any(m.occur(m.ixeqtn{2},m.ixname{3},t),1);
  aux = find(all([aux1;aux2],1));
  if ~isempty(aux)
    error_(6,m.name(1,m.ixname{4}(aux)),'Residual variable in both measurement and transition equation(s): ''%s''.',m.fname);
  end

  end % of nested function ----------------------------------------------------------------------------------

end % of primary function -----------------------------------------------------------------------------------
  
  function [name,label,value] = namelist_(list) % subfunction -----------------------------------------------
  
  patt = '(?<label>".*?")?\s*(?<name>\w+)\s*(?<value>=[^;,]+)?';
  x = regexp(list,patt,'names');
  label = {x(:).label};
  name = {x(:).name};
  value = {x(:).value};
  label = regexprep(label,'"','');
  value = regexprep(value,'=','');

  end % of subfunction --------------------------------------------------------------------------------------

  function invalid = chkkey_(keywords) % subfunction --------------------------------------------------------

  list = {...
    '@variables:transition',...
    '@variables:measurement',...
    '@variables:residual',...
    '@variables:log',...
    '@parameters',...
    '@substitutions',...
    '@equations:transition',...
    '@equations:measurement',...
    '@outside',...
    '@mapoftunes',...
    '@if','@else','@end','@for','@stop','@allbut','@linear',...
    '@log','@exp','@sqrt','@isinf','@isnan','@cumsum','@cumprod','@sum','@prod',...
  };

  invalid = {};
  for i = 1 : length(keywords)
    if ~any(strcmp(keywords{i},list))
      invalid{end+1} = keywords{i};
    end
  end

  end % of subfunction --------------------------------------------------------------------------------------

  function flag = chkallbut_(tokens) % subfunction ----------------------------------------------------------

  % @allbut must be in all or none of log declaration blocks
  index = cellfun(@isempty,regexp(tokens,'@allbut','match','once'));
  if any(index) && any(~index), error_(52); end

  end % of subfunction --------------------------------------------------------------------------------------

  function chksyntax_(m__,nT__,t) % subfunction -------------------------------------------------------------

  x = rand(1,length(m__.name),nT__);
  undeclared__ = {};
  syntax__ = {};
  % eqtn__ = strrep(m__.eqtnF,'t',sprintf('%g',t__));
  eqtn__ = m__.eqtnF;
  for eq__ = 1 : length(eqtn__)
    try
      eval(eqtn__{eq__});
    catch
      errormsg__ = geterrormsg();
      [match__,tokens__] = regexpi(errormsg__,'undefined.*''(\w*)''','match','tokens','once');
      if ~isempty(match__)
        undeclared__{end+1} = tokens__{1};
        undeclared__{end+1} = m__.eqtn{eq__};
      else
        syntax__{end+1} = m__.eqtn{eq__};
        syntax__{end+1} = sprintf('Matlab says: %s',errormsg__);
      end
    end
  end
     
  if ~isempty(undeclared__), error_(38,undeclared__,m__.fname); end
  if ~isempty(syntax__), error_(8,syntax__,m__.fname); end

  end % of subfunction --------------------------------------------------------------------------------------
