function [m,npath,eigval] = solve(m,varargin)
%
% SOLVE  Compute first-order accurate reduced-form solution.
%
% Syntax:
%   [m,npath,eigval] = solve(m,...)
% Arguments:
%   m model; npath logical; eigval numeric
% Options:
%  'alt' logical|numeric|Inf (inf) Selection of parameterizations to be solved for.
%  'reduce' logical (true) Remove duplicit variable occurences from state vector.
%  'forward' numeric (0) Expand model solution forward up to t+k.
%  'persist' logical (false) If true preserve model derivatives as anonymous functions in memory until 'persist' is set to false.
%  'tolerance' numeric (getrealsmall()) Threshold for eigenvalue test.
%
% The IRIS Toolbox 2007/02/03. Copyright 2007 Jaromir Benes.

default = {...
  'alt',inf,...
  'double',true,...
  'reduce',true,...
  'forward',0,...
  'persist',false,...
  'tolerance',getrealsmall(),...
};
options = passopt(default,varargin{1:end});

% function body ---------------------------------------------------------------------------------------------

% no parameterisations
if isempty(options.alt) || (islogical(options.alt) && ~any(options.alt)), return, end

% warning if some parameters are not assigned
[flag,list] = isnan(m,'parameters');
if flag == true, warning_(2,list); end

[m,npath] = solve_(m,options.forward,options.tolerance,options.reduce,options.persist,options.alt);

% no stable solution
index = npath == 0;
if any(index), warning_(26,{sprintf(' %g',find(index))}); end
% infinitely many stable solutions
index = isinf(npath);
if any(index), warning_(27,{sprintf(' %g',find(index))}); end
% complex derivatives
index = isnan(npath);
if any(index), warning_(29,{sprintf(' %g',find(index))}); end
% singularity in state space
index = npath == -1;
if any(index), warning_(32,{sprintf(' %g',find(index))}); end

if nargout > 2, eigval = m.eigval; end

end % of primary function -----------------------------------------------------------------------------------