function [C,Q] = acf(w,varargin)
%
% ACF  Autocovariance and autocorrelation function.
%
% Syntax:
%   [C,R] = acf(w,...)
% Arguments:
%   C numeric; R numeric; w rvar|svar
% Options:
%   'applyto' logical|numeric|Inf (Inf) Index of variables to apply the filter to before ACF is computed
%   'filter' char (empty) Time-domain or frequency-domain filter
%   'nfreq' numeric (256) Number of evenly spaced frequencies to integrate over in the frequency domain when computing filtered ACF
%   'order' numeric (0) Maximum order of autocovariance or autocorrelation computed.
%
% The IRIS Toolbox 2007/02/03. Copyright 2007 Jaromir Benes.

[ny,p,nalt] = dim(w);

default = {
  'applyto',Inf,...
  'filter','',...
  'nfreq',256,...
  'order',0,...
};
options = passopt(default,varargin{:});

if isinf1(options.applyto)
  options.applyto = true([1,ny]);
elseif ~islogical(options.applyto)
  aux = options.applyto;
  options.applyto = false([1,ny]);
  options.applyto(aux) = true;
end

% function body ---------------------------------------------------------------------------------------------

% linear filter applied to some of variables
if ~isempty(options.filter) && any(options.applyto)
  isfilter = true;
  [frq,filter] = fdfilter(options.nfreq,options.filter);
else
  isfilter = false;
end

[T,R,K,Z,H,D,Za,Omega,w] = sspace(w);
for ialt = 1 : nalt 
  if isfilter == true
    % we should multiply by 2*width == 2*pi/nfreq but we skip dividing S by 2*pi in XSFVAR and hence skip multiplying it by 2*pi here
    C(:,:,:,ialt) = real(xsfvar(w.A(:,:,ialt),w.Omega(:,:,ialt),frq,filter,options.applyto,options.order))/options.nfreq;    
  else
    c = acovf(T(:,:,ialt),R(:,:,ialt),[],[],[],[],Za(:,:,ialt),Omega(:,:,ialt),w.eigval(1,:,ialt),0);
    if p > 1, c = reshape(c(1:ny,:),[ny,ny,p]); end
    if p == 0, c(:,:,end+1:options.order+1) = 0;
      elseif options.order > p - 1, c = acovfyw(w.A(:,:,ialt),c,options.order);
      else, c = c(:,:,1:1+options.order); end
    C(:,:,:,ialt) = c;
  end
end

% autocorrelation function
if nargout > 1, Q = cov2corr(C); end

end % of primary function -----------------------------------------------------------------------------------