function varargout = llf(x,range,varargin)
%
% LLF  Local level filter (random walk plus noise) with hard-tune options.
%
% Syntax:
%   [tnd,gap] = llf(x)
%   [tnd,gap] = llf(x,range,...)
% Arguments:
%   tnd tseries; gap tseries; x tseries; range numeric|Inf
% Options:
%   'drift' numeric (0)
%   'lambda' numeric (100*get(x,'freq')^2)
%   'level' tseries (empty)
%   'forecast' numeric (0)
%   'growth' tseries (empty)
%   'swap' logical (false)
%
% The IRIS Toolbox 2007/02/03. Copyright 2007 Jaromir Benes.

freq = datfreq(x.start);
default = {
  'drift',0,...
  'lambda',10*freq,...
  'level',[],...
  'forecast',0,...
  'growth',[],...
  'swap',false,...
};
options = passopt(default,varargin{1:end});

if nargin < 2, range = Inf;
  else range = setrange(range); end

% function body ---------------------------------------------------------------------------------------------

if options.lambda <= 0
  if freq == 0, error('No default smoothing parameter for time series with indeterminate frequency.');
    else, error('Unable to use non-positive smoothing parameter.'); end
end

if isempty(range), [varargout{1:nargout}] = deal(empty(x)); return, end

[y,range] = getdata_(x,range);
[T,n] = size(y);

w = options.lambda;
aux = [-w,2*w+1,-w];
d = aux(ones([1,T]),:);

d(1,2) = w+1;
d(end,2) = w+1;

B = spdiags(d,-1:1,T,T);

yy = y;
yy(1) = yy(1) - w*options.drift;
yy(end) = yy(end) + w*options.drift;
%yy(1) = yy(1) - options.lambda*drift(2);
%yy(2:end-1) = yy(2:end-1) + options.lambda*(drift(2:end-1)-drift(3:end));
%yy(end) = yy(end) + options.lambda*drift(end);  

if isa(options.level,'tseries') && len(options.level) > 0
  if get(options.level,'start') < range(1) || get(options.level,'end') > range(end)
    options.level = resize(options.level,range);
    warning('Level constraint(s) outside filtering range ignored.');
  end
  dates = vech(find(~isnan(options.level)));
  index = 1 + round(dates - range(1));
  yy = [yy;ref(getdata_(options.level,dates),1:length(dates),ones([1,n]))];
  for i = index
    B(end+1,i) = 1;
    B(i,end+1) = 1;
  end
end

if isa(options.growth,'tseries') && len(options.growth) > 0
  if get(options.growth,'start') < range(2) || get(options.growth,'end') > range(end)
    options.growth = resize(options.growth,range(2:end));
    warning('Growth constraint(s) outside filtering range ignored.');
  end
  dates = vech(find(~isnan(options.growth)));
  index = 1 + round(dates - range(1));
  yy = [yy;ref(getdata_(options.growth,dates),1:length(dates),ones([1,n]))];
  for i = index
    B(end+1,[i-1,i]) = [-1,1];
    B([i-1,i],end+1) = [-1;1];
  end
end

tnd = ref(B\yy,1:T,1:n);
gap = y - tnd;

if options.forecast > 0
  for i = 1 : options.forecast
    tnd(end+1,:) = tnd(end,:) + options.drift;
    gap(end+1,:) = 0;
  end
  range = [range,range(end)+(1:options.forecast)];
end

if options.swap == false, [first,second] = deal(tnd,gap);
  else, [first,second] = deal(gap,tnd); end

varargout{1} = x;
varargout{1}.start = range(1);
varargout{1}.data = first;
varargout{1} = cut_(varargout{1});
if nargout > 1
  varargout{2} = x;
  varargout{2}.start = range(1);
  varargout{2}.data = second;
  varargout{2} = cut_(varargout{2});
end

end % of primary function -----------------------------------------------------------------------------------