function dbase = dbload(fname,varargin)
%
% DBLOAD  Load database from CSV file.
%
% db = dbload(fname,...)
% Arguments:
%   db struct, fname char
% Options:
%   'case' char (empty)
%   'datecols' numeric (1)
%   'freq' numeric (NaN)
%   'freqletters' (default)
%   'nan' cellstr|char ({'na','nan','#n/a'})
%
% The IRIS Toolbox 2007/02/03. Copyright 2007 Jaromir Benes.

irisconfig = load('irisconfig.mat');

default = {
  'case','',...
  'changecase','',...
  'datecols',1,...
  'freq',NaN,...
  'freqletters',irisconfig.freqletters,...
  'nan',{'na','nan','#n/a'},...
};
options = passopt(default,varargin{1:end});

if ~isempty(options.changecase) && isempty(options.case), options.case = options.changecase; end

if ~isa(options.freqletters,'char') || length(options.freqletters) < 3, options.freqletters = irisconfig.freqletters; end

if ischar(options.nan), options.nan = {options.nan}; end
options.nan = lower(options.nan);

% -----function DBLOAD body----- %

csvdelim = ',';

[file,newline,firstLine] = readcsv_(fname);

if isempty(file), error('Incorrect format or empty file.'); end

% remove the leading ARCH or % lines if present
line = lower(get1stline_(file,newline));
line(isspace(line)) = '';
while ~isempty(strfind(lower(line),'arch')) ...
  || all(line == ',' | line == ' ' | line == char(10))  ...
  || line(1) == '%'
  [file,newline] = del1stline_(file,newline);
  line = lower(get1stline_(file,newline));
  line(isspace(line)) = '';
end

% read variable names
aLine = get1stline_(file,newline);
[names,active] = readcells_(aLine,true,fname,options);
if strcmpi(options.changecase,'upper'), names = upper(names);
  elseif strcmpi(options.changecase,'lower'), names = lower(names); end
[file,newline] = del1stline_(file,newline);

% search for classes / comments / comment lines
[classes,comments] = deal({});
aLine = get1stline_(file,newline);
firstcell = get1stcell_(aLine);
firstcell = strtrim(strrep(firstcell,'"',''));
donothing = isempty(firstcell) || (firstcell(1) == '%');
isclass = ~donothing && (~isempty(strfind(lower(firstcell),'org.')) || ~isempty(strfind(lower(firstcell),'class')));
iscomment = ~donothing && ~isempty(strfind(lower(firstcell),'comment'));
while (donothing || isclass || iscomment) && ~isempty(file)
  if isclass
    classes = lower(readcells_(aLine,false,fname,options));
  elseif iscomment
    comments = readcells_(aLine,false,fname,options);
  end
  [file,newline] = del1stline_(file,newline);
  aLine = get1stline_(file,newline);
  firstcell = get1stcell_(aLine);
  firstcell = strtrim(strrep(firstcell,'"',''));
  donothing = isempty(firstcell) || (firstcell(1) == '%');
  isclass = ~donothing && ~isempty(strfind(lower(firstcell),'org.')) || ~isempty(strfind(lower(firstcell),'class'));
  iscomment = ~donothing && ~isempty(strfind(lower(firstcell),'comment'));
end
classes(end+1:length(active)) = {''};
comments(end+1:length(active)) = {''};

if isempty(file)
  
  range = NaN;
  data = NaN([1,length(active)]);

else

  file(strfind(file,'"')) = '';
  file(file <= char(32) & file ~= char(10)) = '';
  file = lower(file);
  firstcell = lower(firstcell);
  options.freqletters = lower(options.freqletters);

  % replace all variants of NaN
  index = strcmp(options.nan,'nan');
  options.nan(index) = [];
  for i = 1 : length(options.nan)
    file = strrep(file,options.nan{i},'nan');
  end

  if options.datecols == 1  
  	startdate = str2dat(firstcell,options.freqletters);
  	[year,per,freq] = dat2ypf(startdate);
    index = find(freq == [1,2,4,12]);
    if ~isempty(index), letter = options.freqletters(index);
      else letter = '?'; end
    % split date column
    file(strfind(file,letter)) = csvdelim;
    file = strrep(file,'nan?nan','nan,nan');
  end

  file(file ~= ',' & file ~= '.' & (file < '0' | file > '9') ...
    & file ~= char(10) & file ~= '+' & file ~= '-' & ...
    file ~= 'e' & file ~= 'E' ) = '';
  
  x = dlmread_(file,csvdelim);

  sample = find(all(~isnan(x(:,1:2)),2));
  
  if options.datecols == 2
    if isnan(options.freq)
      freq = autofreq_(x(sample,2));
    else
      freq = options.freq;
    end
  end
  range = vech(datcode(x(sample,1),x(sample,2),freq));
  data = x(sample,3:end);

end

dataLen = size(data,1);
      
% create database
dbase = struct();
invalid = cell([1,0]);
for k = vech(find(active));
  if isempty(classes{k}) || ~isempty(strfind(classes{k},'ts'))
    try, dbase.(names{k}) = tseries(range,data(:,k),comments{k});
      catch, invalid(end+1:end+2) = {upper(fname),names{k}}; end
  elseif ~isempty(classes{k}) && ~isempty(strfind(classes{k},'double'))
    pos = strfind(classes{k},'double');
    arrayDim = vech(sscanf(classes{k}(pos+6:end),'[%g]'));  
    try, dbase.(names{k}) = reshape(data(1:prod(arrayDim),k),arrayDim);
      catch, invalid(end+1:end+2) = {upper(fname),names{k}}; end
  end
end

if ~isempty(invalid), warning_(2,invalid); end

end % of primary function -----------------------------------------------------------------------------------

  function [names,active] = readcells_(line,check,fname,options) % subfunction ------------------------------

  names = {};
  active = [];

  names = regexp(line,',"([^"]*)"|,([^,]*)','tokens');
  names  = strtrim([names{:}]);
  
  try, active = ~cellfun(@isempty,names);
    catch, active = ~cellfun('isempty',names); end

  if check == true
    invalid = regexp(names,'[a-zA-Z]\w*','match','once');
    index = ~strcmp(names,invalid);
    if any(index)
      aux = cell([1,2*sum(index)]);
      aux(1:2:end) = {upper(fname)};
      aux(2:2:end) = names(index);
      error_(3,aux);
    end
  end

  end % of subfunction --------------------------------------------------------------------------------------

  function cell = get1stcell_(line) % subfunction -----------------------------------------------------------
  
  if isempty(line)
    cell = '';
  else
    x = regexp(line,'(?<cell>.*?)(,|$)','names','once');
    cell = strtrim(x.cell);
  end
  
  end % of subfunction --------------------------------------------------------------------------------------

  function line = get1stline_(file,ixnewline)

  if isempty(ixnewline), line = '';
    else, line = strtrim(file(1:ixnewline(1))); end

  end % of subfunction --------------------------------------------------------------------------------------

  function [file,newlines] = del1stline_(file,newlines)

  if isempty(newlines)
    file = [];
    newlines = [];
  else
    len = newlines(1);
    file = file(newlines(1)+1:end);
    newlines = newlines(2:end) - len;
  end

  end % of subfunction --------------------------------------------------------------------------------------
  
  function [out_file,out_newline,out_firstLine] = readcsv_(fname)
  
  % reads text file into string variable and cuts leading and trailing blank lines
  fid = fopen(fname,'r');
  if fid == -1
    msg = iff(exist(fname,'file'),'Unable to open ''%s'' for reading.','Unable to find ''%s'' in current folder.');
    error(msg,upper(fname));
  end

  out_file = transpose(char(fread(fid)));
  if isempty(out_file) || out_file(end) ~= char(10)
    out_file = [out_file,char(10)];
  end
  out_file(find(out_file == char(13))) = [];
  out_newline = find(out_file == char(10));

  if fclose(fid) == -1
    error(['Unable to close ''',upper(fname),''' after reading.']);
  end

  letter = '0';
  
  u = find(out_file >= letter);
  if isempty(u)
    out_firstLine = 0;
    lastline      = 0;
  else
    out_firstLine = whichline_(out_newline,u(1));
    lastline = whichline_(out_newline,u(end));
  end
  
  nfirst = 1 : out_firstLine-1;
  nlast  = lastline+1 : length(out_newline);
  
  for k = nfirst
    [out_file,out_newline] = del1stline_(out_file,out_newline);
  end
  n = length(out_newline);
  for k = nlast
    [out_file,out_newline] = deleol_(out_file,out_newline);
  end
  
  end % of subfunction --------------------------------------------------------------------------------------

  function [out_line] = whichline_(in_newline,in_position)
  
  u = find(in_position < in_newline);
  if isempty(u)
    out_line = length(in_newline) + 1;
  else
    out_line = u(1);
  end
  
  end % of subfunction --------------------------------------------------------------------------------------

  function [inout_file,inout_newline] = deleol_(inout_file,inout_newline)
  
  if isempty(inout_newline)
    inout_file = [];
    inout_newline = [];
  else
    inout_file = inout_file(1:inout_newline(end));
    inout_newline = inout_newline(1:end-1);
  end
  
  end % of subfunction --------------------------------------------------------------------------------------

function freq = autofreq_(x)

maxX = max(x);
if maxX > 4
  freq = 12;
elseif maxX > 2
  freq = 4;
elseif maxX > 1
  freq = 2;
else
  freq = 1;
end

end