function notsaved = dbsave(d,fname,range,varargin)
%
% DBSAVE  Save database to CSV file.
%
% Syntax:
%   notsaved = dbsave(db,fname)
%   notsaved = dbsave(db,fname,rng,...)
% Arguments:
%   notsaved cellstr; db struct|dbase; fname char; rng numeric
% Options:
%  'class' logical (true)
%  'comment' logical (true)
%  'decimal' numerice (empty)
%  'format' char ('%.8e')
%  'freqletters' char (irisconfig.freqletters)
%  'nan' char ('NaN')
%  'arch' logical (false)
%
% The IRIS Toolbox 2007/02/03. Copyright 2007 Jaromir Benes.

config = load('irisconfig.mat');

default = {
  'arch',false,...
  'class',true,...
  'comment',true,...
  'decimal',[],...
  'format','%.8e',...
  'freqletters',config.freqletters,...
  'nan','NaN',...
  'troll',false,...
};
options = passopt(default,varargin{:});

if ~isempty(options.decimal)
  options.format = ['%.',sprintf('%g',options.decimal),'e'];
end

if ~isa(options.freqletters,'char') || length(options.freqletters) < 3
  options.freqletters = default{3,2};
end

% function body ---------------------------------------------------------------------------------------------

if nargin < 3 || isinf1(range), range = 'max';
  else, range = vech(range); end

tseriesname = {};
numericname = {};
notsaved = {};
list = fieldnames(d);
for i = 1 : length(list)
  if istseries(d.(list{i}))
    tseriesname(end+1) = list(i);
  elseif isnumeric(d.(list{i})) || islogical(d.(list{i})) || ischar(d.(list{i}))
    numericname(end+1) = list(i);
  else
    notsaved(end+1) = list(i);
  end
end

% ntseries = length(tseriesname);
nnumeric = length(numericname);

name = {};
dim = {};
clss = {};
cmt = {};
% tseries to numeric array conversion
if ~isempty(tseriesname)
  array = tseries([],[]);;
  for i = 1 : length(tseriesname)
    x = d.(tseriesname{i});
    si = size(x);
    si = [si(1),prod(si(2:end))];
    x = reshape(x,si);
    try
      array = [array,x];
    catch
      notsaved(end+1) = tseriesname(i);
      continue
    end
    dim(end+1) = {si(2)};
    name(end+1) = tseriesname(i);
    clss(end+1) = {'tseries'};
    name(end+(1:si(2)-1)) = {''};
    dim(end+(1:si(2)-1)) = {[]};
    clss(end+(1:si(2)-1)) = {''};
    if si(2) == 1, cmt(end+1) = {get(x,'comment')};
      else, cmt(end+(1:si(2))) = get(x,'comment'); end
  end
  [data,range] = double(array,range);
else
  data = [];
  range = zeros([1,0]);
end

if ~isempty(notsaved), warning_(6,notsaved); end

% numeric
for i = 1 : length(numericname)
  x = d.(numericname{i});
  si = size(x);
  dim(end+1) = {si};
  difflen = prod(si) - length(range);
  if difflen > 0
    range(end+(1:difflen)) = nan([1,difflen]);
    data(end+(1:difflen),:) = nan([difflen,size(data,2)]);
  end
  data(:,end+1) = NaN;
  data(1:prod(si),end) = double(x(:));
  name(end+1) = numericname(i);
  clss(end+1) = {class(x)};
  cmt(end+1) = {''};
end

fid = fopen(fname,'w+');
if fid == -1, error(['Unable to open ''',upper(fname),''' for writing.']); end

% write leading ARCH row if troll format desired
if options.arch == true || options.troll == true
  aux = cell([1,length(name)-1]);
  aux(1:end) = {''};
  fwrite(fid,sprintf('ARCH%s\n',sprintf(',%s',aux{:})),'char');
end

% write names
fwrite(fid,sprintf('%s\n',sprintf(',%s',name{:})),'char');

% write comments
if options.comment == true
  cmt = regexprep(cmt,'(.*)','"$0"');
  fwrite(fid,sprintf('Comment%s\n',sprintf(',%s',cmt{:})),'char');
end

% write classes and sizes
if options.class == true && options.arch == false && options.troll == false
  aux = 'Class';
  for i = 1 : length(name)
    aux = [aux,sprintf(',%s',clss{i})];
    if ~isempty(dim{i}), aux = [aux,sprintf('[%g]',dim{i})]; end
  end
  aux = [aux,sprintf('\n')];
  fwrite(fid,aux,'char');
end

% date format string
[year,per,freq] = dat2ypf(range);
if isempty(freq)
  dateformat = '%g';
  year(:,1) = [];
else
  index = find(freq(1) == [1,2,4,12]);
  if isempty(index)
    dateformat = '%g';
    year(:,1) = [];
  else
    options.freqletter = strrep(options.freqletters,',',':');
    options.freqletter = strrep(options.freqletters,';',':');
    dateformat = ['%g',options.freqletters(index),'%g'];
  end  
end

% data format string
if size(data,2) > 0
  aux = cell([1,size(data,2)]);
  aux(1:end) = {options.format};
  dataformat = sprintf(',%s',aux{:});
end
formatline = [dateformat,dataformat,sprintf('\n')];
data = [year(:),per(:),data];
aux = sprintf(formatline,transpose(data));
if ~strcmp(lower(options.nan),'nan'), aux = strrep(aux,'NaN',options.nan); end
aux = strrep(aux,sprintf(dateformat,NaN,NaN),'NaN');
fwrite(fid,aux,'char');

if fclose(fid) == -1, warning(['Unable to close ''',upper(fname),''' after writing.']); end

end % of primary function -----------------------------------------------------------------------------------