function x = compile(inputfile,outputfile,varargin)
%
% COMPILE  Compile DVI, PS or PDF file from TEX, DVI or PS file.
%
% Syntax:
%   compile(inputfile,outputfile,...)
% Arguments:
%   inputfile,outputfile char
% Options:
%   'display' logical (true) Display or not output messages from LaTeX, DVIPS and/or DVIPDFM.
%   'orientation' 'portrait'|'landscape' ('portrait') Paper orientation.
%   'papersize' 'a4'|'letter'|'executive' ('a4') Paper size.
%   'ps2pdf' logical (false) If true PDF files are always created from PS files.
%
% The IRIS Toolbox 2007/02/13. Copyright 2007 Jaromir Benes.

default = {...
  'deletetex',true,...
  'display',true,...
  'papersize','a4',...
  'ps2pdf',false,...
  'orientation','portrait',...
};
options = passopt(default,varargin{:});

% function body ---------------------------------------------------------------------------------------------

irisconfig = load('irisconfig.mat');
if isempty(irisconfig.latexpath), error('LaTeX not linked. Unable to use COMPILE.'); end

% check paper size
options.papersize = strrep(options.papersize,'paper',''); 
list = {'a4','letter','executive'};
if all(~strcmp(options.papersize,list)), options.papersize = 'a4'; end

% check paper orientation
list = {'portrait','landscape'};
if all(~strcmp(options.orientation,list)), options.orientation = 'portrait'; end

% tempfile to catch latex.exe output if display == false
tempfile = tempname;

[inputpath,inputtitle,inputext] = fileparts(inputfile);
[outputpath,outputtitle,outputext] = fileparts(outputfile);

% check output extension
outputext = lower(outputext);
list = {'.ps','.pdf','.dvi'};
if all(~strcmp(outputext,list)), outputext = '.ps'; end

% suppress output display if desired
if options.display == true
  quiet = '';
elseif strcmp(outputext,'.ps')
  quiet = '-q*';
else
  quiet = '-q';
end

% create DVI
if strcmp(inputext,'.tex')
  system([irisconfig.latexpath,' ',inputfile,iff(options.display,'',sprintf('>> %s',tempfile))]);
  if options.display == false, delete(tempfile); end
end
if ~strcmp(inputtitle,outputtitle), movefile([inputtitle,'.dvi'],[outputtitle,'.dvi']); end

% create PS
if ~strcmp(inputext,'.ps') && (strcmp(outputext,'.ps') || (strcmp(outputext,'.pdf') && options.ps2pdf == true))
  systemstring = sprintf('%s -t %s %s %s %s.dvi -o %s.ps',irisconfig.dvipspath,options.papersize,iff(strcmp(options.orientation,'landscape'),'-t landscape',''),quiet,outputtitle,outputtitle);
  system(systemstring);
end

% create PDF
if strcmp(outputext,'.pdf')
  if strcmp(inputext,'.ps') || options.ps2pdf == true
    systemstring = sprintf('%s %s.ps',irisconfig.ps2pdfpath,outputtitle);
    system(systemstring);
  else
    % PDF output
    systemstring = sprintf('%s -p %s %s %s %s.dvi',irisconfig.dvipdfmpath,options.papersize,iff(strcmp(options.orientation,'landscape'),'-l',''),quiet,outputtitle);
    system(systemstring);
  end
end

% move output file to desired folder
if ~isempty(outputpath), flag = movefile(sprintf('%s%s',outputtitle,outputext),outputpath); end

end % of primary function -----------------------------------------------------------------------------------